/** \file mlc8k.c Performs USB Mass Storage operations for 8k-page MLC FLASH.
    Requires the correct initializations performed, see mlcinit.c .
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vs1000.h>

#include <audio.h>

#include <player.h>
#include <mappertiny.h>
#include <minifat.h>
#include <vsNand.h>

#include <dev1000.h>

#include "mlc8k.h"

//#define COPYBACK_WRITE /*Read for copy-back 0x35*/

/*
  MLC-8kB 45ns 2GB
  write            : 420-460kB/sec (windows copy)
  write w/ copyback: 460-510kB/sec (windows copy)
  read             : 467kB/sec (cygwin diff)

  MLC 80ns 81620MB (FAT not counted)
  write: 125kB/sec  (cygwin copy) 360kB/sec (windows copy)
  read : 658kB/sec
  MLC 45ns 81620MB (FAT not counted)
  write: 125kB/sec
  read : 680kB/sec
  SLC
  write: 480kB/sec
  read:  670kB/sec
 */

//#define PATCH_SCSI_COMMAND

#ifdef PATCH_SCSI_COMMAND
#include <scsi.h>
int startUnit = 1;
#endif


#ifdef COPYBACK_WRITE
#include "vsnand.h"
extern struct FsNandPhys fsNandPhys;
u_int16 NandGetStatus(void);
s_int16 Large8kPageCopyBackWrite(u_int32 page) {
  PERIP(GPIO0_SET_MASK) = GPIO0_CS1; //NandDeselect(); /* For robustness */
  NandSetWaits(fsNandPhys.waitns);
  PERIP(GPIO0_CLEAR_MASK) = GPIO0_CS1;//  NandSelect();

//  NandPutCommand(NAND_OP_READ_A);/*Sets read/write pointer*/
  NandPutCommand(NAND_OP_COPY_BACK_PROGRAM);
  NandPut8kDataAddress(page & ~15L); /* not valid if low 4 bits set! */

/* TODO: bytes can be patched if necessary.. */

  NandPutCommand(NAND_OP_PERFORM_PROGRAM); //Program the data and spare
  {
      register u_int16 n = NandGetStatus();
      PERIP(GPIO0_SET_MASK) = GPIO0_CS1;//      NandDeselect();

#if DEBUG_LEVEL != 0
      if (n & 1) {
	  puthex(n);
	  puts("=COPYBACK");
      }
#endif
      return (n & 1); //(n & 1) ? (s_int16)n : 0; //Error code from flash      
  }
}
#endif


/*
  Only large-page FLASH supported.
*/

/*
  6kW X:0x0000..0x17ff (fmfCache).
  9kW Y:0x1c00..0x3fff (treeBuild)
  2kW X:0x2400..0x27ff (start of MallocAreaX, USBMem in X:0x2800..0x3400)
  2kW Y:0x1000..0x17ff (dcthi)
*/

#define USE_USB
#define USE_MAP

#define DEBUG_LEVEL 0

void InitMlc(void);

#if DEBUG_LEVEL != 0
void puthex6(u_int32 a) {
    puthex(a>>16);
    puthex(a);
}
#endif

int IsValid(u_int32 sector);

extern struct FsPhysical *ph;
extern struct FsMapper *map;
//struct FsMapper *oldMap;


#define NandSelect()   PERIP(GPIO0_CLEAR_MASK) = GPIO0_CS1
#define NandDeselect() PERIP(GPIO0_SET_MASK) = GPIO0_CS1
void NandPutDataAddress(register __c u_int32 addr);
u_int16 NandGetStatus(void);
extern struct FsNandPhys fsNandPhys;


#if DEBUG_LEVEL > 2
void PrintData(u_int16 *d, u_int16 l) {
    int i;
    for(i=0;i<l;i++) {
	puthex(*d++);
	if ((i&7) == 7)
	    puts("");
    }
}
#endif


#ifdef USE_MAP
struct FsMapper *FsMapMlcCreate(struct FsPhysical *physical,
				u_int16 cacheSize);
u_int16 FsMapMlcRead(struct FsMapper *map, u_int32 firstBlock,
		     u_int16 blocks, u_int16 *data);
u_int16 FsMapMlcWrite(struct FsMapper *map, u_int32 firstBlock,
		      u_int16 blocks, u_int16 *data);
u_int16 FsMapMlcFlush(struct FsMapper *map, u_int16 flush);

/*const*/ struct FsMapper mlcMapper = {
    0x010c, /*version*/
    256,    /*blocksize in words*/
    0,      /*blocks -- read from CSD*/
    0,      /*cacheBlocks*/
    NULL,//FsMapMlcCreate,
    FsMapFlNullOk,//RamMapperDelete,
    FsMapMlcRead,
    FsMapMlcWrite,
    NULL,//FsMapFlNullOk,//RamMapperFree,
    FsMapMlcFlush,//RamMapperFlush,
    NULL /* no physical */
};

//extern __y u_int16 treeBuild[CACHE_SECTORS][256]; /* cache data */
//extern __y u_int16 g_dcthi[2048];

u_int32 GetPhysical(u_int32 logical) {
#if 1
    register u_int16 bl = logical / ph->eraseBlockSize;
    register int i = 0;
    while (i < mlc.bads) { /* 15 cycles per loop */
	if (mlc.bad[i] > bl)
	    break;
	bl++;
	i++;
    }
#if DEBUG_LEVEL > 2
    puthex6(logical);
    putchar('>');
    puthex6((logical % ph->eraseBlockSize) + (u_int32)bl * ph->eraseBlockSize);
    puts("");
#endif
    return (logical & (ph->eraseBlockSize-1)) + (u_int32)bl * ph->eraseBlockSize;
#else
    return logical;
#endif
}

#ifdef DIRECT_WRITE
int IsValid(u_int32 sector) {
    /* Always read the spare directly.
       Also works optimally with the physical cache
       (if it isn't written yet, it can be modified)! */
    u_int16 meta[8];
    ph->Read(ph, sector, 1, NULL, meta);
#if 1
    if ((meta[1] & 0xff) != 0xff /*==0x05*/)
	return 1;
#if DEBUG_LEVEL > 0
    if ((meta[0] & meta[1] & meta[2] & meta[3] & meta[4] & meta[5]) != 0xffffU) {
	puts("\033[1m");
	puthex(meta[0]);
	puthex(meta[1]);
	puthex(meta[2]);
	puthex(meta[3]);
	puthex(meta[4]);
	puthex(meta[5]);
	puts("=M\033[0m");
    }
#endif
    return 0;
#else
    return (meta[1] & 0xff) == 0x05;
#endif
}
#endif

#if 1
u_int16 GetBitArray(register __i0 u_int16 *array, register __a0 int bit);
void SetBitArray(register __i0 u_int16 *array, register __a0 int bit);
void ClearBitArray(register __i0 u_int16 *array, register __a0 int bit);
#else
u_int16 GetBitArray(register __i0 u_int16 *array, register __a0 int bit) {
    register __a1 int t = array[bit >> 4] & (1<<(bit&15));
    return t;
}
void SetBitArray(register __i0 u_int16 *array, register __a0 int bit) {
    array[bit >> 4] |= (1<<(bit&15));
}
void ClearBitArray(register __i0 u_int16 *array, register __a0 int bit) {
    array[bit >> 4] &= ~(1<<(bit&15));
}
#endif

#if 0
struct REPL *GetReplWith(u_int16 block) {
    register int i;
    for (i=0; i<mlc.replaced; i++)
	if ((u_int32)block * ph->eraseBlockSize == mlc.replace[i].new)
	    return &mlc.replace[i];
    return NULL;
}
#endif

void Merge(struct REPL *r);

/* Get the next unused erase block from the replace area. */
u_int16 GetBlock(void) {
    static u_int16 cur = 0;
 re:
    if (cur >= ph->eraseBlocks || cur < (ph->eraseBlocks-RESERVED_AREA))
	cur = (ph->eraseBlocks-RESERVED_AREA);

#if 0
    if (GetReplWith(cur)) {
	cur++;
	goto re;
    }
#else
    {
	register int i;
	for (i=0; i<mlc.replaced; i++) {
	    if ((u_int32)cur * ph->eraseBlockSize == mlc.replace[i].new) {
		cur++;
		goto re;
	    }
	}
    }
#endif

#if 1 /* 20110520: Should skip bad erase blocks! */
    {
	u_int16 t[8];
	memset(t, -1, 16/2);
	Large8kPageReadSingle(ph, (u_int32)cur*ph->eraseBlockSize, 1, NULL, t);
	if ((t[0+2] & 0xff) != 0xff) {
	    cur++;
	    goto re;
	}
    }
#endif

#if DEBUG_LEVEL > 1
    puthex(cur);
    puts(" alloc");
#endif
#if 1 /* Clear it to be certain during debugging. -- does not really slow down. */
    if (ph->Erase(ph, (u_int32)cur * ph->eraseBlockSize >> 2)) {
	//putch('E');
	//putch('1');
    }
#endif
#if 0
    puthex(cur);
    puts("bl");
#endif
    return cur++;
}

#if 0
struct REPL *PutReplaced(u_int16 block, u_int16 repl) {
    register struct REPL *r = &mlc.replace[mlc.replaced];
    __y u_int16 *d = r->data;
    memset(r, 0, sizeof(mlc.replace[0]));
    r->data = d;
    //r->bits = 0;
    //r->time = 0;
    r->orig = (u_int32)block * ph->eraseBlockSize;
    r->new  = (u_int32)repl * ph->eraseBlockSize;
    //r->wrPtr = 0;
    replaced++;
    return r;
}
#endif

#if 0
void RemReplaced(struct REPL *r) {
#if 0
    register int i;
    for (i=0; i<mlc.replaced; i++) {
	if (r == &mlc.replace[i]) {
	    __y u_int16 *d = r->data;
#if DEBUG_LEVEL > 1
	    puthex(i);
	    puthex(mlc.replaced);
	    puthex(mlc.replace[i].orig);
	    puthex(mlc.replace[i].new);
	    puts(" RemReplaced");
#endif
	    memcpy(&mlc.replace[i], &mlc.replace[i+1],
		   sizeof(mlc.replace[0])*(MAXREPL-i));
	    mlc.replaced--;
	    mlc.replace[MAXREPL-1].data = d; /* don't lose the buffer pointer */
	    return;
	}
    }
#else
    memcpy(r, r+1, (char *)&mlc.replace[MAXREPL-1]-(char *)r);
    mlc.replaced--;
#endif
}
#else
#define RemReplaced(r) do{register __y u_int16 *d = (r)->data;mlc.replaced--;memcpy((r), (r)+1, (char *)&mlc.replace[MAXREPL-1]-(char *)(r));mlc.replace[MAXREPL-1].data=d;}while(0)
#endif

struct REPL *GetReplaced(u_int16 block) {
    register int i;
    for (i=0; i<mlc.replaced; i++)
	if ((u_int32)block * ph->eraseBlockSize == mlc.replace[i].orig)
	    return &mlc.replace[i];
    return NULL;
}


u_int16 dataTmp[PAGESIZE / 2];
__y u_int16 mapTmp[MAX_ERASEBLOCKSIZE/SECTORSPERPAGE];

auto void Fill(struct REPL *r) {
    register u_int32 p;
    u_int16 meta[SECTORSPERPAGE/*16*/ *8];
    if (r->mask == 65535U/*65535, 15*/)
	return;
    if (GetBitArray(r->replbitmap, r->se>>SECSPERPAGEBITS)) {
	register int i;
	for (i=r->wrPtr-SECTORSPERPAGE/*16*/;i>=0;i-=SECTORSPERPAGE/*16*/) {
	    p = r->new + i;
	    ph->Read(ph, p, 1, NULL, meta);
	    if ((meta[4] & (ph->eraseBlockSize-1)) == r->se) {
		/* Found! */
		goto journal;
	    }
	}
#if 0
	puts("fill: no journal");
#endif
    }
    p = r->phys;
 journal:
#if 0
    puthex(r->sector>>16);
    puthex(r->sector);
    //putchar(' ');
    puthex(p>>16);
    puthex(p);
#endif
    /* Use datatmp -- more robust than reading one at a time. */
#ifdef COPYBACK_WRITE
    Large8kPageReadForCopyBack(p, dataTmp, meta);
#else
    Large8kPageRead(p, dataTmp, meta);
#endif
    {
	register int i;
	for (i=0;i<SECTORSPERPAGE;i++) {
	    if ((((s_int16)r->mask>>i) & 1) == 0) {
		memcpyXY(r->data+i*512/2, dataTmp+i*512/2, 512/2);
#ifdef DIRECT_DEBUG
		puthex(meta[3]);
		puthex(i);
		puts(" f");
#endif
	    }
	}
    }
    r->mask = 65535U;
}

auto void SetMeta(register __c s_int32 sector, register __i2 u_int16 *meta) {
    memset(meta, 0xffff, SECTORSPERPAGE*8*sizeof(u_int16));
    meta++;
    *meta = 0xff05;       //[1]
    meta += 2;
    *meta++ = sector;     //[3]
    *meta++ = sector;     //[4]
    *meta++ = sector>>16; //[5]
    *meta++ = sector;     //[6]
    meta += 2;
    {
	register int i;
	for (i=0;i<SECTORSPERPAGE-1;i++) {
	    *meta++ = 0xff05;     //[1+n*8]
	    meta += 7;
	}
    }
}

void Merge(struct REPL *r) {
    /* First get an extra erase block (does not really allocate it). */
    register int i;
    register u_int16 bl = 0;
    int cnt = 0;

PERIP(GPIO1_CLEAR_MASK) = LED1;

#if 0
    puthex(r->orig>>16);
    puthex(r->orig);
    puts("Merge");
#endif

    bl = GetBlock();

#if DEBUG_LEVEL > 0
    puthex(r->orig);
    putchar(',');
    puthex(r->new);
    puts(" orig,new");

    for (i=0;i<256/16;i++) {
	puthex(r->replbitmap[i]);
    }
    puts("=replbitmap");
#endif

    if (r->mask) {
	Fill(r);
	SetBitArray(r->replbitmap, r->se>>SECSPERPAGEBITS);
    }

    /* Copy all untouched blocks from r->orig to bl. */
    for (i=0; i<ph->eraseBlockSize; i+=SECTORSPERPAGE/*16*/) {
	u_int16 meta[SECTORSPERPAGE/*16*/*8];
	if ( GetBitArray(r->replbitmap, i>>SECSPERPAGEBITS) ) {
	    /* replaced data */
	} else {
	    /* copy data that was not overwritten */
#ifdef COPYBACK_WRITE
	    /*TODO: */
	    Large8kPageReadForCopyBack(r->orig +i, dataTmp, meta);
#else
	    Large8kPageRead(r->orig +i, dataTmp, meta);
#endif

#if 0
	    if (meta[4] != 0xffff &&
		*(u_int32*)&meta[4] != r->logical+i) {
		putch(0xaa);
		putch(0);
		putch(meta[5]>>8);
		putch(meta[5]);
		putch(meta[4]>>8);
		putch(meta[4]);
		putch(i>>8);
		putch(i);
	    }
#endif
#if 0
	    if (meta[4] != 0xffff &&
		(meta[4] & (ph->eraseBlockSize-1)) != i) {
		putch(meta[4]>>8);
		putch(meta[4]);
		putch(i>>8);
		putch(i);
	    }
#endif
	    //meta[1] = 0xff05;
#ifdef COPYBACK_WRITE
	    //remove this..
	    //SetMeta(r->orig + i, meta); /*was this correct? was r->logical*/
#else
	    SetMeta(r->logical + i, meta); /*is this correct?*/
#endif
	    //SetMeta(r->logical + i, meta); /*Is this correct? r->orig?*/
	    if (Large8kPageWrite((u_int32)bl*ph->eraseBlockSize+cnt, dataTmp, meta)) {
		//putch('W');
		//putch('1');
		/*TODO: Erase block replacement! */
	    }
	    cnt += SECTORSPERPAGE/*16*/;
	}
    }
    /* Erase r->orig. */
    if (ph->Erase(ph, r->orig >> 2)) {
	//putch('E');
	//putch('2');
    }

    memsetY(mapTmp, 0, sizeof(mapTmp));
    for (i=0; i<r->wrPtr;i+=SECTORSPERPAGE/*16*/) {
	u_int16 meta[8];
    reread:
	ph->Read(ph, r->new + i, 1, NULL, meta);
	/* Read error? */
#if 0
	if (meta[4] != meta[6] || meta[4] != meta[3]) {
	    putch(i>>8);
	    putch(i);
	    putch(meta[4]>>8);
	    putch(meta[4]);
	    putch(meta[5]>>8);
	    putch(meta[5]);
	    putch(meta[6]>>8);
	    putch(meta[6]);
	    putch(meta[3]>>8);
	    putch(meta[3]);
	    goto reread;
	}
#endif
	if (CountBitsLong(meta[4] ^ meta[6])
	    + CountBitsLong(meta[4] ^ meta[3]) > 2) {
#if 0
	    putch('r');
	    putch('0' + CountBitsLong(meta[4] ^ meta[6]) + CountBitsLong(meta[4] ^ meta[3]));
#endif
	    goto reread;
	}
	/*
	  a b c Majority vote
	  1 1 1
	  1 1 0
	  1 0 1
	  0 1 1
	  (a&b) | (a&c) | (b&c) = (a & (b|c)) | (b&c)
	 */
	{
	    register u_int16 ll =
		(meta[4] & (meta[6] | meta[3])) | (meta[6] & meta[3]);
	    ll &= (ph->eraseBlockSize-1);
	    mapTmp[ll >> SECSPERPAGEBITS] = i;
	}
    }


    {
	register int j;
	i = 0;
	for (j=0; j<ph->eraseBlockSize; j+=SECTORSPERPAGE/*16*/) {
	    u_int16 meta[SECTORSPERPAGE/*16*/*8];
#ifdef COPYBACK_WRITE
	    __y u_int16 copyback = 0;
#endif
	    if (GetBitArray(r->replbitmap, j>>SECSPERPAGEBITS)) {
		/* touched, either in memory or in journal block */
		if (r->mask && r->se == j) {
		    memcpyYX(dataTmp, r->data, PAGESIZE / 2);
		    SetMeta(r->sector, meta);
		} else {
		    /*Find the page from journal.
		      Use a cache for speedup. */
#ifdef COPYBACK_WRITE
		    if (Large8kPageReadForCopyBack(r->new + mapTmp[j>>SECSPERPAGEBITS],
					dataTmp, meta) == 0) {
			copyback = 1;
		    }
#else
//		    int re = 4;
//		rdagain1:
		    if (Large8kPageRead(r->new + mapTmp[j>>SECSPERPAGEBITS],
					dataTmp, meta) & 0x7fff) {
#if 0
			putch('R');
			putch('1');
#endif
//			if (re-- > 0)
//			    goto rdagain1;
		    }
#endif
#if 0
		    if (meta[4] != 0xffff &&
			*(u_int32*)&meta[4] != r->logical+j) {
#if 0
			putch(0xaa);
			putch(1);
			putch(meta[5]>>8);
			putch(meta[5]);
			putch(meta[4]>>8);
			putch(meta[4]);
			putch(j>>8);
			putch(j);
#else
			putch('N');
			//putch('0'+(j/1600)%10);
			putch('0'+(j/160)%10);
			putch('0'+(j/16)%10);
#endif
		    }
#endif
		}
	    } else {
		/* not touched, get the next saved page */
#ifdef COPYBACK_WRITE
		if (Large8kPageReadForCopyBack((u_int32)bl * ph->eraseBlockSize + i,
				    dataTmp, meta) == 0) {
		    copyback = 1;
		}
#else
//		int re = 4;
//	    rdagain2:
		if (Large8kPageRead((u_int32)bl * ph->eraseBlockSize + i,
				    dataTmp, meta) & 0x7fff) {
#if 0
		    putch('R');
		    putch('2');
#endif
//		    if (re-- > 0)
//			goto rdagain2;
		}
#endif
#if 0
		if (meta[4] != 0xffff &&
		    *(u_int32*)&meta[4] != r->logical+j) {
		    putch(0xaa);
		    putch(2);
		    putch(meta[5]>>8);
		    putch(meta[5]);
		    putch(meta[4]>>8);
		    putch(meta[4]);
		    putch(j>>8);
		    putch(j);
		}
#endif
		i += SECTORSPERPAGE/*16*/;
	    }
#ifndef COPYBACK_WRITE /*for Hynix*/
	    SetMeta(r->logical + j, meta);
#endif
	    /* TODO: for latter 2 cases use copy-back program,
	       but 8kPageRead needs to report if ECC was applied! */
#ifdef COPYBACK_WRITE
	    if (copyback &&
		Large8kPageCopyBackWrite(r->orig + j) == 0) {
	    } else
#endif
	    if (Large8kPageWrite(r->orig + j, dataTmp, meta)) {
		//putch('W');
		//putch('5');
		/*TODO: Erase block replacement! */
	    }
	}
    }
    /* Erase r->new */
    if (ph->Erase(ph, r->new >> 2)) {
	//putch('E');
	//putch('3');
    }

    /* Erase bl. */
    if (ph->Erase(ph, (u_int32)bl * ph->eraseBlockSize >> 2)) {
	//putch('E');
	//putch('4');
    }

    RemReplaced(r);
PERIP(GPIO1_SET_MASK) = LED1;
}


u_int16 FsMapMlcRead(struct FsMapper *map, u_int32 sector,
		     u_int16 blocks, u_int16 *data) {
    u_int16 bc = 0;
#ifdef USE_USB
    while (bc < blocks) {
	register u_int32 phy = GetPhysical(mlc.dataStart + sector);
	register u_int16 se = phy & (ph->eraseBlockSize-1);
	u_int16 bl = phy / ph->eraseBlockSize;
	register struct REPL *r = GetReplaced(bl);
	u_int16 meta[8];
#if 0
	puthex(sector);
	puts("=R");
#endif


#if DEBUG_LEVEL > 1
	//puthex6(sector);
	//putchar(' ');
	puthex6(phy);
	puts(" RD");
#endif
	if (r) {
	    /* Replaced, check if a new version exists. */
#if 0 //DEBUG_LEVEL > 1
	    puts("r");
#endif
	    if (r->se == (se & ~(SECTORSPERPAGE-1))
		&& (r->mask & (1U<<(se & (SECTORSPERPAGE-1))))) {
		/* Got it in cache. */
#if 0
		puthex(phy>>16);
		puthex(phy);
		puts(" ch");
#endif
#if DEBUG_LEVEL > 2
		printf(" read ch %06lx %d\n", phy, r->data[(phy&(SECTORSPERPAGE-1))*8+7]);
#endif
		memcpyYX(data, &r->data[(phy & (SECTORSPERPAGE-1))*512/2],
			 512/2*sizeof(u_int16));
		meta[1] = 0xff05;
		goto noread;
	    } else if ( GetBitArray(r->replbitmap, se>>SECSPERPAGEBITS) ) {
		/* a new version, try to locate it. */
		register int i;
#if DEBUG_LEVEL > 1
		puts(" yes");
#endif
		for (i=r->wrPtr-SECTORSPERPAGE/*16*/; i>=0; i-=SECTORSPERPAGE/*16*/) {
		    register u_int16 ll;
		    ph->Read(ph, r->new + i, 1, NULL, meta);
		    ll = (meta[4] & (meta[6] | meta[3])) | (meta[6] & meta[3]);

		    if ((se & ~(SECTORSPERPAGE-1)) == (ll & (ph->eraseBlockSize-1))) {
			phy = r->new + i + (se&(SECTORSPERPAGE-1));
#if DEBUG_LEVEL > 1
			//puthex(r->wrPtr-i);
			//putchar(' ');
			puthex6(phy);
			puts(" repl");
#endif
#if 0
			//puthex(phy);
			puts("j");
#endif
			break;
		    }
		}
#if 0
		if (i<0) {
		  puthex(phy>>16);
		  puthex(phy);
		  puts(" no j!");
		}
#endif
		/* Error, replbitmap invalid */
	    } else {
		/* Not replaced, use original */
	    }
	} else {
	    /* Not replaced, use original if valid.. */
	}
	ph->Read(ph, phy, 1, data, meta);
    noread:
	if ((meta[1] & 0xff) != 0x05) {
	    /* Not a data block. Sector not available or never written. */
	    /* Return ZERO data! */
#if DEBUG_LEVEL > 0
	    puthex(sector);
	    puts("NA");
#endif
#if 0
	    puts("NA");
#endif
	    memset(data, 0, 512/2);
#if 0
	} else if (meta[4] == (u_int16)sector &&
		   meta[5] == (u_int16)(sector>>16)) {
	    /* data ok */
	} else {
#if DEBUG_LEVEL > 0
	    puthex(meta[5]);
	    puthex(meta[4]);
	    putchar('!');
	    puthex(sector);
	    puts("");
#endif
	    /* Sector not found at the right place. */
#endif
	}
gotit:
	data += 512/2;
	sector++;
	bc++;
    }
#endif
    return bc;
}
u_int16 FsMapMlcWrite(struct FsMapper *map, u_int32 sector,
		      u_int16 blocks, u_int16 *data) {
    u_int16 bc = 0;
    while (bc < blocks) {
	register u_int32 phy = GetPhysical(mlc.dataStart + sector);
	struct REPL *r;
	u_int16 se;
#if 0
	puthex(sector);
	puts("=W");
#endif
#if DEBUG_LEVEL > 1
	puthex6(sector);
	puts(" WR");
#endif
//	data[1] = 0xff05;
//	data[4] = sector;
//	data[5] = sector>>16;

    re:
	se = phy % ph->eraseBlockSize;
	r = GetReplaced(phy / ph->eraseBlockSize);
	if (!r) {
	    /*AllocReplace*/
	    if (mlc.replaced >= MAXREPL) { /* if all slots used */
		register int i;
		register u_int16 idx = 0, high = 0U;
#if DEBUG_LEVEL > 0
		puts("Out of replace space, force flush MERGE");
#endif
		for (i=0; i<mlc.replaced; i++) {
		    register u_int16 b = mlc.replace[i].bits + mlc.replace[i].time *1;
		    if (high < b) {
			high = b;
			idx = i;
		    }
		}
#if 0
		putch('M');
		putch('1');
#endif
		Merge(&mlc.replace[idx]);
	    }
	    {
		register int i;
		for (i=0;i<MAXREPL;i++)
		    if (mlc.replace[i].time < 0xffffU)
			mlc.replace[i].time++;
	    }
	    //r = PutReplaced(phy / ph->eraseBlockSize, GetBlock());
	    r = &mlc.replace[mlc.replaced];
	    {
		register __y u_int16 *d = r->data;
		memset(r, 0, sizeof(*r));
		r->data = d;
	    }
	    r->inOrder = 1; /* ordered by default */
	    r->logical = sector & ~(u_int32)(ph->eraseBlockSize-1);
	    r->orig    = phy    & ~(u_int32)(ph->eraseBlockSize-1);
	    r->new     = (u_int32)GetBlock() * ph->eraseBlockSize;
	    mlc.replaced++;
	}
	if (r->se != (se & ~(SECTORSPERPAGE-1)) && r->mask) {
	    /* find the unwritten blocks from r->old or journal */
	    u_int16 meta[SECTORSPERPAGE*8];
	    Fill(r);
#if DEBUG_LEVEL > 0
	    //printf(" %04x %06lx\n", r->orig, r->sector);
#endif
#ifdef DIRECT_WRITE
	    if (r->wrPtr == 0 && /* Stop writing if any block goes to journal. */
		r->se == r->wrDirect && (r->se != 0 || !IsValid(r->orig))) {
		/* Write to the original sector directly. */
		//putch('D');
		SetMeta(r->sector, meta);
#if 1
		memcpyYX(dataTmp, r->data, PAGESIZE/2);
		Large8kPageWrite(r->phy/*??*/r->sector, dataTmp, meta);
#else
		Large8kPageWrite(r->phy/*??*/r->sector, r->data, meta);
#endif
		r->wrDirect += SECTORSPERPAGE;
	    } else
#endif /*DIRECT_WRITE*/
	    {
		if (r->wrPtr >= ph->eraseBlockSize) {
#if DEBUG_LEVEL > 0
		    puts(" Journal for block full!");
#endif
#if 0
		    putch('M');
		    putch('2');
#endif
		    Merge(r);
		    goto re;
		}
		/* Write to journal */
		SetMeta(r->sector, meta);
#if 1
		memcpyYX(dataTmp, r->data, PAGESIZE/2);
		if (Large8kPageWrite(r->new + r->wrPtr, dataTmp, meta)) {
		    //putch('W');
		    //putch('6');
		    /*TODO: Erase block replacement! */
		}
#else
		if (Large8kPageWrite(r->new + r->wrPtr, r->data, meta)) {
		    //putch('W');
		    //putch('6');
		    /*TODO: Erase block replacement! */
		}
#endif
		if (r->se != r->wrPtr)
		    r->inOrder = 0; /* no longer ordered */

		r->wrPtr += SECTORSPERPAGE;
		/* replbitmap can be reduced to 1/4 */
		if (!GetBitArray(r->replbitmap, r->se>>SECSPERPAGEBITS))
		    r->bits++;
		SetBitArray(r->replbitmap, r->se>>SECSPERPAGEBITS);
	    }
	    r->mask = 0;
	}
	/* the same 8k */
#if DEBUG_LEVEL > 0
	//printf("->c %06lx %d\n", phy, r->bits);
#endif
	r->sector = sector & ~(SECTORSPERPAGE-1L);
	r->phys = phy & ~(SECTORSPERPAGE-1L);
	r->se = se & ~(SECTORSPERPAGE-1);
//Fill(r);
	r->mask |= (1U << (se & (SECTORSPERPAGE-1)));
	memcpyXY(&r->data[(se & (SECTORSPERPAGE-1))*512/2], data, sizeof(u_int16)*512/2);

/*SHOULD BE ACTIVE! Speeds up by flushing "terminated" blocks. */
	if (r->mask == 65535U /*15*/ &&
	    r->se + SECTORSPERPAGE == ph->eraseBlockSize ) {
#ifdef DIRECT_WRITE
	    if (r->wrPtr == 0 && r->se == r->wrDirect) {
		/* Write to the original sector */
		u_int16 meta[SECTORSPERPAGE*8];
		//putch('F');
		SetMeta(r->sector, meta);
#if 1
		memcpyYX(dataTmp, r->data, PAGESIZE/2);
		Large8kPageWrite(r->sector, dataTmp, meta);
#else
		Large8kPageWrite(r->sector, r->data, meta);
#endif
		RemReplaced(r);
	    } else
#endif /*DIRECT_WRITE*/
	    {
#if DEBUG_LEVEL > 0
		puts("MERGE filled");
#endif
#if 0
		putch('M');
		putch('3');
#endif
		Merge(r);
	    }
	}
	data += 512/2;
	sector++;
	bc++;
    }
    return bc;
}
u_int16 FsMapMlcFlush(struct FsMapper *map, u_int16 hard) {
#if 0
    puts("=F");
#endif
#if DEBUG_LEVEL > 0
    puthex(replaced);
    puts("Flush()");
#endif
    /*
      Note: this slows down writes by cygwin cp considerably: 272->121kB/sec.
     */
    while (mlc.replaced) {
#if 0
	putch('M');
	putch('4');
#endif
	Merge(&mlc.replace[0]);
    }
    return 1;
}
#if 0
/* In checkmlc.c */
struct FsMapper *FsMapMlcCreate(struct FsPhysical *physical,
				u_int16 cacheSize) {
    /* Scan erase blocks to locate the first data block,
       bad blocks, and possible replaced blocks. */
    ...
    return &mlcMapper;
}
#endif

#endif/*USE_MAP*/



#ifdef USE_USB
#include <usblowlib.h>
#endif/*USE_USB*/

#ifdef PATCH_SCSI_COMMAND
auto u_int16 NewDiskProtocolCommand(register __i2 u_int16 *cmd) {
#if 1
    if ((cmd[OPERATION_CODE] & 0xff) == SCSI_START_STOP_UNIT) {
	/* 0:OP 1:imm0 2:res 3:res 4:power7-4 res3-2 loej1 start0 5:control */
	startUnit = cmd[4] & 3;
    } else if (startUnit == 2) {
	/* The next command after STOP UNIT will not be serviced
	   fully because we will turn ourselves off. */
	startUnit = 4;
    } else
#endif
#if 0
    if ((cmd[OPERATION_CODE] & 0xff) == SCSI_INQUIRY &&
	(cmd[1/*flags__pageCode*/] & 0x01ff) == 0x0180) {
	//Extended page: Serial Number page requested, See SCSI2 chapter 8.3.4.5
	/*Periph qual,type | page code, Res | Len of sernum in bytes, "10001005" */
	static const __y u_int16 testSerNumber[] = {
	    0x0080, 0x0008, 0x3130, 0x3030, 0x3130, 0x3035
	};
	SCSI.DataOutBuffer = SCSI.DataBuffer;
	//SCSI.Status = SCSI_OK;
	SCSI.State = SCSI_DATA_TO_HOST;
	memcpyYX(SCSI.DataBuffer, testSerNumber, 6);
	SCSI.DataOutSize = 12;             
	return 0; /* handled */
    }
#endif
    return cmd[OPERATION_CODE]; /* perform the command */
}
#endif


void mlc_structure_too_large_error(void);
void main(void) {
    //ph = FsPhNandCreate(1);
//    int i;

#pragma msg 36 off
    if (sizeof(mlc) > 512)
	mlc_structure_too_large_error();
#pragma msg 36 on

#if 0
    clockX = 7;
    SetRate(100);
#endif

    /* Set the leds after nand-boot! */
    PERIP(GPIO1_SET_MASK) =  LED1;      /* POWER led on */
    PERIP(GPIO1_DDR)   |=  LED1|LED2; /* SI and SO to outputs */
    PERIP(GPIO1_MODE)  &= ~(LED1|LED2); /* SI and SO to GPIO */

    ph = FsPhNandCreate(1); /* Init the default physical first. */
    ph->Read = Large8kPageReadSingle; /*Note: prog change fixed*/
//fsNandPhys.waitns = 50;
#ifdef USE_MAP
#if 0
    map = FsMapMlcCreate(ph, 0); /* Then create mapper v2. */
#else
    map = &mlcMapper; //FsMapMlcCreate(ph, 0); /* Then create mapper v2. */
    mlcMapper.blocks = mlc.dataBlocks;
#endif
#else
    map = FsMapTnCreate(ph, 0); /* tiny mapper */
#endif

#ifdef USE_USB
//    while(1)
    if (USBIsAttached()) {
#if DEBUG_LEVEL > 0
	puts("Going to MassStorage");
#endif
	PERIP(GPIO1_SET_MASK) = LED2;      /* feature LED on */
#ifdef PATCH_SCSI_COMMAND
	SetHookFunction((u_int16)MSCPacketFromPC, PatchDiskProtocolCommand);
#endif
	/*
	  We replace the MassStorage() functionality.
	*/
	voltages[voltIoUSB] = 31; /* 3.65V */
	voltages[voltAnaUSB] = 31; /* 3.65V */
	PowerSetVoltages(&voltages[voltCoreUSB]);
	/** a 2.5ms-10ms delay to let the voltages change (rise) */
	BusyWait10();
	LoadCheck(NULL, 1); /* 4.0x clock required! */
	InitUSB(USB_MASS_STORAGE);  

#if 1 /* Change serial number */
#define D_SIZE 26
	{
	    static const u_int16 myStringDescriptor3[] = {
		//String 2:
		(D_SIZE<<8)|03, //type=string
		('3'<<8)|0, ('0'<<8)|0, ('0'<<8)|0, ('0'<<8)|0,
		('1'<<8)|0, ('0'<<8)|0, ('0'<<8)|0, ('0'<<8)|0,
		('1'<<8)|0, ('0'<<8)|0, ('0'<<8)|0, ('0'<<8)|0
	    };
#undef D_SIZE
	    USB.descriptorTable[3] = myStringDescriptor3;
	}
#endif

	while (1) {
	    USBHandler();

	    /* If no SOF in such a long time.. */
	    if (USBWantsSuspend()) {
		/* Do we seem to be detached?
		   Note: Now also requires USBWantsSuspend() to return TRUE! */
		if (USBIsDetached()) {
		    break;
		}
#if 0 /* Suspend */
		{
		    register u_int16 io = PERIP(GPIO1_ODATA);
		    PERIP(GPIO1_ODATA) = io & ~(LED1|LED2);      /* LEDs off */
		    PERIP(SCI_STATUS) |= SCISTF_ANA_PDOWN;       /*fix suspend*/
		    USBSuspend(0); /* no timeout */
		    PERIP(GPIO1_ODATA) = io;      /* LEDs back to old states */
		    /* with USB audio the LED settings are handled by user interface */
		}
		USB.lastSofTime = ReadTimeCount();
#endif
	    }
#ifdef PATCH_SCSI_COMMAND
	    if (startUnit == 4) {
		break;
	    }
#endif
	}
	PERIP(SCI_STATUS) &= ~SCISTF_USB_PULLUP_ENA; /* Take us 'out of the bus' */
	PERIP(USB_CONFIG) = 0x8000U; /* Reset USB */
	map->Flush(map, 2);

#ifdef PATCH_SCSI_COMMAND
	if (startUnit == 4) {
	    PowerOff();
	}
#endif

#if 0
	/* Clean unused FAT sectors. */
	if (map->Free && FatInitFileSystem() == 0) {
	    FatIterateOverFreeSectors(map->Free, map);
	}
#endif
	/* Set player voltages. No need to wait for the voltages to drop. */
	LoadCheck(NULL, 0);
	PowerSetVoltages(&voltages[voltCorePlayer]);
#ifdef PATCH_SCSI_COMMAND
	// SetHookFunction((u_int16)MSCPacketFromPC, RealMSCPacketFromPC);
#endif
	PERIP(GPIO1_CLEAR_MASK) = LED2;      /* LED off */
#if 0
	/* Clear mapping in case we have got more bad blocks */
	if (mlc.badsChanged) {
	    ph->Erase(ph, ((ph->eraseBlockSize < 256) ? 256 : ph->eraseBlockSize) >> 2);
	}
#endif
    }
#endif
//    while (1)
	;
#ifdef USE_MAP
    map = FsMapTnCreate(ph, 0); /* tiny mapper */
#endif


#if 1
    while (1) {
#define PROG_START 0 //12288 //8192
#if 0
	ph->Read(ph, PROG_START/512, 16, mallocAreaX, minifatBuffer);
#else
	{
	    register int i;
	    for(i=0;i<16;i++) {
		Large8kPageReadSingle(ph, (PROG_START>>9) + i, 1,
				      mallocAreaX+i*512/2, minifatBuffer);
	    }
	}
#endif
	BootFromX(mallocAreaX+8);
    }
#endif
}
