/** \file mlc8k.c Performs USB Mass Storage initializations.
    Performs the correct initializations for mlc8k.c .
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vs1000.h>

#include <player.h>
#include <mappertiny.h>
#include <minifat.h>
#include <vsNand.h>
#include <audio.h>

#include <dev1000.h>

#include "mlc8k.h"

/*
  Only large-page FLASH supported.
*/


extern struct FsPhysical *ph;
extern struct FsMapper *map;

//extern struct Codec *cod;
//extern struct CodecServices cs;

u_int32 GetPhysical(u_int32 logical) {
    register u_int16 bl = logical / ph->eraseBlockSize;
    register int i = 0;
    while (i < mlc.bads) { /* 15 cycles per loop */
	if (mlc.bad[i] > bl)
	    break;
	bl++;
	i++;
    }
    return (logical & (ph->eraseBlockSize-1)) + (u_int32)bl * ph->eraseBlockSize;
}


struct FsMapper *FsMapMlcCreate(struct FsPhysical *physical,
				u_int16 cacheSize) {
    /* Scan erase blocks to locate the first data block,
       bad blocks, and possible replaced blocks. */
    register u_int16 eB, eraseBlockSize = physical->eraseBlockSize;
    register u_int32 sector = 0;

    memset(&mlc, 0, sizeof(mlc));
    //mlc.bads = 0;
    mlc.rootBlock =
	((physical->eraseBlockSize < 0x100) ? 0x100 : physical->eraseBlockSize)
	+ physical->eraseBlockSize /*reserve one*/;

    if (sizeof(mlc) > 533/*615*/) {
#pragma msg 36 off
	error("sizeof(mlc) > 533\n");
#pragma msg 36 on
    }

    {/*init data buffer pointers, 11 words*/
	register int i;
	register struct REPL *r = &mlc.replace[0];
	register __y u_int16 *p = mallocAreaY;
	for (i=0;i<MAXREPL;i++) {
	    r->data = p;
	    p += PAGESIZE/2; //was 2048/2;
	    r++;
	}
    }

    for (eB = 0; eB < physical->eraseBlocks; eB++) {
#if 0
	physical->Read(ph, (u_int32)eB * eraseBlockSize, 2, NULL, (void *)tmpBuf);
#else
	physical->Read(ph, (u_int32)eB * eraseBlockSize, 1, NULL, (void *)tmpBuf);
	physical->Read(ph, (u_int32)eB * eraseBlockSize+1, 1, NULL, (void *)&tmpBuf[8]);
#endif
	if ((tmpBuf[0+2] & tmpBuf[8+2] & 0xff) != 0xff) {
	    mlc.bad[mlc.bads++] = eB;
	    if (mlc.bads == MAX_BADS) {
		/* too many BAD's */
		break;
	    }
	} else if ((tmpBuf[1] & 0xff) == 0x05) {
	    /* Erase block has been written to at least once. */

	    /* Datastart must be found before other data sectors. */
	    if (*(u_int32 *)&tmpBuf[4] == 0 && mlc.dataStart == 0) {
		/* logical-to-physical offset */
		mlc.dataStart = sector;
	    }
	}
	sector += eraseBlockSize;
    }
    /* leave room for mapper.v1 structure. */
    mlc.dataStart = GetPhysical(mlc.rootBlock + physical->eraseBlockSize);
    mlc.dataBlocks = (u_int32)(physical->eraseBlocks - RESERVED_AREA - mlc.bads)
	* eraseBlockSize - mlc.dataStart;
    return NULL; //&mlcMapper;
}


void main(void) {
    /* Set the leds after nand-boot! */
    PERIP(GPIO1_ODATA) |=  LED1;      /* POWER led on */
    PERIP(GPIO1_DDR)   |=  LED1|LED2; /* SI and SO to outputs */
    PERIP(GPIO1_MODE)  &= ~(LED1|LED2); /* SI and SO to GPIO */

    ph = FsPhNandCreate(0);
    /*Replace all.*/
    ph->Read = Large8kPageReadSingle; /*Note: prog change fixed*/
    FsMapMlcCreate(ph, 0);

#if 1
    /* Check that mapper-compatible structure is valid and upto date. */
    if (((struct FsNandPhys *)ph)->nandType < 6) {
	/* Compare structure and rewrite it if necessary. */
	__y u_int32 phys = GetPhysical(mlc.rootBlock);
	u_int16 i, err = 0;
	for (i=0;i<ph->eraseBlocks;i++) {
	    u_int32 sec = GetPhysical(mlc.dataStart +
				      (u_int32)i * ph->eraseBlockSize);
	    if ((i & 4095) == 0) {
		Large8kPageRead(phys + ((i>>12)<<4),
				mallocAreaX, minifatBuffer);
	    }
	    if (mallocAreaX[i & 4095] == sec / ph->eraseBlockSize) {
		/* correct */
	    } else {
		/* incorrect, update is required. */
		err++;
		break;/* ignore rest */
	    }
	}
	if (!err) {
	    /*puts("ok");*/
	} else {
	    /* Update mapping structure */
	    phys = GetPhysical(mlc.rootBlock);
	    ph->Erase(ph, phys>>2);
	    memset(mallocAreaX, -1, 2048/2);
	    for (i=0;i<ph->eraseBlocks;i++) {
		u_int32 sec = GetPhysical(mlc.dataStart +
					  (u_int32)i * ph->eraseBlockSize);
		mallocAreaX[i & 4095] = sec / ph->eraseBlockSize;
		if ((i & 4095) == 4095 || i == ph->eraseBlocks-1) {
		    memset(minifatBuffer, -1, 16*16/2);
		    minifatBuffer[1] = 0xff01;
		    Large8kPageWrite(phys + ((i>>12)<<4),
				     mallocAreaX, minifatBuffer);
		    memset(mallocAreaX, -1, 2048/2);
		}
	    }
	}
    }


#else
    if (((struct FsNandPhys *)ph)->nandType < 6) {
	/* check that nandType is sensible */

//putch(mlc.rootBlock>>24);
//putch(mlc.rootBlock>>16);
//putch(mlc.rootBlock>>8);
//putch(mlc.rootBlock);
//putch('r');

#define DO_ALWAYS
#ifndef DO_ALWAYS
	/* Check if Mapper would be already satisfied. */
        if (ph->Read(ph, mlc.rootBlock, 1/*4*/, NULL, (void *)tmpBuf) == 1
	    && (tmpBuf[1] & 0xff) == 0x01) {
	    /*puts("MAP Ok already!");*/
	} else
#endif
	{
	    __y u_int32 phys = /*0x3ffe00; */ GetPhysical(mlc.rootBlock);
	    u_int16 i;
//putch('I');
	    ph->Erase(ph, phys>>2);
	    memset(mallocAreaX, -1, 2048/2);
	    for (i=0;i<ph->eraseBlocks;i++) {
		u_int32 sec = GetPhysical(mlc.dataStart +
					  (u_int32)i * ph->eraseBlockSize);
		mallocAreaX[i & 4095] = sec / ph->eraseBlockSize;
		if ((i & 4095) == 4095 || i == ph->eraseBlocks-1) {
		    memset(minifatBuffer, -1, 16*16/2);
		    minifatBuffer[1] = 0xff01;
//putch(phys>>24);
//putch(phys>>16);
//putch(phys>>8);
//putch(phys);
//putch('W');
		    Large8kPageWrite(phys + ((i>>12)<<4),
				     mallocAreaX, minifatBuffer);
		    memset(mallocAreaX, -1, 2048/2);
		}
	    }
	}
    }
#endif
    while (1) {
	register u_int32 start = PLAYERSTART; /* player init */
	if (USBIsAttached()) {
	    start = MLCSTART; /* USB: MLC mapper */

	    /*
	      Perform some other setup required by the MLC code
	      to save code space on mlc.c.
	     */
	    SetRate(44100U);
	    //PERIP(GPIO1_SET_MASK) =  LED2;
	    voltages[voltIoUSB] = 31; /* 3.65V */
	    voltages[voltAnaUSB] = 31; /* 3.65V */
	    PowerSetVoltages(&voltages[voltCoreUSB]);
	    /** a 2.5ms-10ms delay to let the voltages change (rise) */
	    BusyWait10();
	}
#if 0
	ph->Read(ph, start >> 9, 16/*max*/, mallocAreaX, minifatBuffer);
#else
	{
	    register int i;
	    for(i=0;i<16;i++) {
		Large8kPageReadSingle(ph, (start>>9) + i, 1,
				      mallocAreaX+i*512/2, minifatBuffer);
	    }
	}
#endif
	BootFromX(mallocAreaX+8);
	/* should not return */
    }
}
